/*
 * @brief HID generic example
 *
 * @note
 * Copyright(C) NXP Semiconductors, 2013
 * All rights reserved.
 *
 * @par
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * LPC products.  This software is supplied "AS IS" without any warranties of
 * any kind, and NXP Semiconductors and its licensor disclaim any and
 * all warranties, express or implied, including all implied warranties of
 * merchantability, fitness for a particular purpose and non-infringement of
 * intellectual property rights.  NXP Semiconductors assumes no responsibility
 * or liability for the use of the software, conveys no license or rights under any
 * patent, copyright, mask work right, or any other intellectual property rights in
 * or to any products. NXP Semiconductors reserves the right to make changes
 * in the software without notification. NXP Semiconductors also makes no
 * representation or warranty that such application will be suitable for the
 * specified use without further testing or modification.
 *
 * @par
 * Permission to use, copy, modify, and distribute this software and its
 * documentation is hereby granted, under NXP Semiconductors' and its
 * licensor's relevant copyrights in the software, without fee, provided that it
 * is used in conjunction with NXP Semiconductors microcontrollers.  This
 * copyright, permission, and disclaimer notice must appear in all copies of
 * this code.
 */

#include <stdio.h>
#include <string.h>
#include "app_usbd_cfg.h"
#include "fsl_device_registers.h"

#include "board.h"
#include "fsl_debug_console.h"
#include "hid_generic.h"
#include "romapi_5411x.h"
#include "fsl_ctimer.h"

#include "pin_mux.h"
#include <stdbool.h>
#include "fsl_power.h"
#include "fsl_fro_calib.h"
/*******************************************************************************
 * Definitions
 ******************************************************************************/

#define CTIMER CTIMER0
/*******************************************************************************
 * Prototypes
 ******************************************************************************/

/* USB descriptor arrays defined *_desc.c file */
extern const uint8_t USB_DeviceDescriptor[];
extern uint8_t USB_HsConfigDescriptor[];
extern uint8_t USB_FsConfigDescriptor[];
extern const uint8_t USB_StringDescriptor[];
extern const uint8_t USB_DeviceQualifier[];
/*******************************************************************************
 * Variables
 ******************************************************************************/
ALIGN(2048) uint8_t g_memUsbStack[USB_STACK_MEM_SIZE];

static USBD_HANDLE_T g_hUsb;

const USBD_API_T *g_pUsbApi;
/*******************************************************************************
 * Code
 ******************************************************************************/

/**
 * @brief	Handle interrupt from USB
 * @return	Nothing
 */
void USB0_IRQHandler(void)
{
	uint32_t *addr = (uint32_t *)USB0->EPLISTSTART;
	/* For EP0 transfers, HW will do auto handshake as long as the ACTIVE is set in
	EP0_IN/OUT command list. Unlike for other endpoints HW will not clear the ACTIVE bit
	after transfer is done. Thus SW should manually clear the bit whenever it receives
	new setup packet and set it only after it has queued the data for control transfer as
	shown in “Flowchart for control endpoint EP0” in the user manual. EP0 OUT ACTIVE bit
	is handled inside the USBD ROM driver already, only EP0 IN ACTIVE bit needs to be cleared. */
	if (USB0->DEVCMDSTAT & _BIT(8))
	{                           /* if setup packet is received */
		addr[2] &= ~(0x80000000); /* clear active bit for EP0_IN */
	}

	USBD_API->hw->EnableEvent(g_hUsb, 0, USB_EVT_SOF, 1);
	USBD_API->hw->ISR(g_hUsb);
}

/**
 * @brief	Find the address of interface descriptor for given class type.
 * @return	If found returns the address of requested interface else returns NULL.
 */
USB_INTERFACE_DESCRIPTOR *find_IntfDesc(const uint8_t *pDesc, uint32_t intfClass)
{
	USB_COMMON_DESCRIPTOR *pD;
	USB_INTERFACE_DESCRIPTOR *pIntfDesc = 0;
	uint32_t next_desc_adr;

	pD = (USB_COMMON_DESCRIPTOR *) pDesc;
	next_desc_adr = (uint32_t) pDesc;

	while (pD->bLength) {
		/* is it interface descriptor */
		if (pD->bDescriptorType == USB_INTERFACE_DESCRIPTOR_TYPE) {

			pIntfDesc = (USB_INTERFACE_DESCRIPTOR *) pD;
			/* did we find the right interface descriptor */
			if (pIntfDesc->bInterfaceClass == intfClass) {
				break;
			}
		}
		pIntfDesc = 0;
		next_desc_adr = (uint32_t) pD + pD->bLength;
		pD = (USB_COMMON_DESCRIPTOR *) next_desc_adr;
	}

	return pIntfDesc;
}

/**
 * @brief	main routine for USB device example
 * @return	Function should not exit.
 */
int main(void)
{
	uint32_t timerFreq;
	ctimer_config_t config;
	USBD_API_INIT_PARAM_T usb_param;
	USB_CORE_DESCS_T desc;
	ErrorCode_t ret = LPC_OK;
	/* Define the init structure for the output LED pin*/
	gpio_pin_config_t led_config = {
		kGPIO_DigitalOutput, 1,
	};

	/* attach 12 MHz clock to FLEXCOMM0 (debug console) */
	CLOCK_AttachClk(BOARD_DEBUG_UART_CLK_ATTACH);

	/* Enable the asynchronous bridge */
	SYSCON->ASYNCAPBCTRL = 1;
	/* Use main clock for ctimer3/4 */
	CLOCK_AttachClk(kMAIN_CLK_to_ASYNC_APB);

	BOARD_InitPins();
	BOARD_BootClockFROHF96M();
	BOARD_InitDebugConsole();
	POWER_DisablePD(kPDRUNCFG_PD_USB0_PHY); /*Turn on USB Phy */
	/* enable USB IP clock */
	CLOCK_EnableUsbfs0Clock(kCLOCK_UsbSrcFro, CLOCK_GetFreq(kCLOCK_FroHf));

	/* Attach Main Clock as CLKOUT */
	CLOCK_AttachClk(kMAIN_CLK_to_CLKOUT);
	if (CLOCK_GetFreq(kCLOCK_FroHf) == 96000000U) {
		/* Set the clock dividor to divide by 8*/
		CLOCK_SetClkDiv(kCLOCK_DivClkOut, 8, false);
	} else {
		/* Set the clock dividor to divide by 4*/
		CLOCK_SetClkDiv(kCLOCK_DivClkOut, 4, false);
	}

	/* Init output LED GPIO. */
	GPIO_PortInit(GPIO, BOARD_LED_RED_GPIO_PORT);
	GPIO_PinInit(GPIO, BOARD_LED_RED_GPIO_PORT, BOARD_LED_RED_GPIO_PIN, &led_config);

	/* Initialize ctimer */
	CTIMER_GetDefaultConfig(&config);
	CTIMER_Init(CTIMER, &config);
	/* Get ctimer clock frequency */
	if ((CTIMER == CTIMER0) || (CTIMER == CTIMER1) || (CTIMER == CTIMER2)) {
		timerFreq = CLOCK_GetFreq(kCLOCK_BusClk);
	} else {
		timerFreq = CLOCK_GetAsyncApbClkFreq();
	}

	/* Return the version of the FRO Calibration library */
	if (fro_calib_Get_Lib_Ver() == 0) {
		while (1);
	}

	/* pass ctimer instance & ctimer clock frquency in KHz */
	ret = Chip_TIMER_Instance_Freq(CTIMER, timerFreq/1000);
	if (ret != LPC_OK) {
		while (1);
	}

	/* initialize USBD ROM API pointer. */
	g_pUsbApi = (const USBD_API_T *) LPC_ROM_API->usbdApiBase;

	/* initialize call back structures */
	memset((void *) &usb_param, 0, sizeof(USBD_API_INIT_PARAM_T));
	usb_param.usb_reg_base = USB0_BASE;
	usb_param.max_num_ep = 2;
	usb_param.mem_base = (uint32_t)&g_memUsbStack;
	usb_param.mem_size = USB_STACK_MEM_SIZE;
	usb_param.USB_SOF_Event = USB_SOF_Event;

	/* Set the USB descriptors */
	desc.device_desc = (uint8_t *) USB_DeviceDescriptor;
	desc.string_desc = (uint8_t *) USB_StringDescriptor;

	/* Note, to pass USBCV test full-speed only devices should have both
	 * descriptor arrays point to same location and device_qualifier set
	 * to 0.
	 */
	desc.high_speed_desc = USB_FsConfigDescriptor;
	desc.full_speed_desc = USB_FsConfigDescriptor;
	desc.device_qualifier = 0;

	/* USB Initialization */
	ret = USBD_API->hw->Init(&g_hUsb, &desc, &usb_param);
	if (ret == LPC_OK) {
		ret =
			usb_hid_init(g_hUsb,
						 (USB_INTERFACE_DESCRIPTOR *) &USB_FsConfigDescriptor[sizeof(USB_CONFIGURATION_DESCRIPTOR)],
						 &usb_param.mem_base, &usb_param.mem_size);
		if (ret == LPC_OK) {
			/*  enable USB interrupts */
			NVIC_EnableIRQ(USB0_IRQn);
			/* now connect */
			USBD_API->hw->Connect(g_hUsb, 1);
		}

	}

	while (1U) {
		/* Sleep until next IRQ happens */
		__WFI();
	}
}
